package nl.tudelft.ewi.alg.distsim;

//  Copyright 2013 Léon Planken and James Boerkoel

/*
    This file is part of the DistSim package.

    DistSim is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DistSim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DistSim.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * <p>Subclasses of this class implement the distributed algorithm.
 * All work should be done in the {@link #receiveMessage(int, Object) receiveMessage} method.</p>
 * 
 * <p>To indicate that the algorithm has work of its own to do, without needing to
 * be activated by a message, let {@link #isDone() isDone} return {@code false}.
 * Then, {@code receiveMessage} will be called with a {@code null} message
 * if there is indeed no incoming message, allowing this work to be done.</p>
 * 
 * <p>Whenever the algorithm would do a non-blocking receive, return from
 * {@code receiveMessage} with {@code isDone} returning {@code false}.
 * To instead do a blocking receive, have {@code isDone} return {@code true}.
 * In this latter case, {@code receiveMessage} will only be called
 * if there is an actual message incoming.</p>
 * 
 * @param <M> The message payload to be sent around.  Can be any type.
 * 
 * @author L&eacute;on Planken
 * @author James Boerkoel Jr.
 */

public abstract class DProcess<M> {
	public final int id;
	protected final MasterProcess<M> masterProc;
	private long nsTotal = 0, msTotal = 0;
	private long nsIdle = 0;
	
	static enum State {
		CREATED(true), RECEIVING(false), WAITING(true), DONE(false), SENDING(false);
		
		final boolean allowDelivery;
		
		State(boolean d) {
			this.allowDelivery = d;
		}
	}
	private State state = State.CREATED;
	
	protected DProcess(MasterProcess<M> master, int id) {
		this.id = master.register(this, id);
		assert this.id == id;
		this.masterProc = master;
	}
	
	void deliverMessage(Message<M> m) {
		assert m.receiver == this;
		if ( !state.allowDelivery ) throw new IllegalStateException("a message is already being delivered");
		state = State.RECEIVING;
		
		if ( m.deliveryTime > getClock() ) {
			if ( !isDone() ) throw new IllegalStateException("this process still had work to do before this message arrived");
			setClock(m.deliveryTime);
		}
		
		final int senderID = m.sender == null ? MasterProcess.DUMMY_PROC_ID : m.sender.id;
		startTimer();
		
		
		receiveMessage(senderID, m.payload);
		
		stopTimer();
		state = State.WAITING;
	}
	
	
	protected void setClock(long value) {
		assert value > getClock() : "can only skip time";
		assert !timerRunning : "can't set clock while timer is running";
		this.nsIdle += value - getClock();
		this.nsTotal = value;
	}

	public long getClock() {
		if ( timerRunning ) {
			stopTimer(); startTimer();
		}
		return nsTotal;
	}
	public long getIdle() {
		return nsIdle;
	}
	
	private long nsStart, msStart;
	private boolean timerRunning = false;
	private void startTimer() {
		assert !timerRunning;
		timerRunning = true;
		msStart = System.currentTimeMillis();
		nsStart = System.nanoTime();
	}
	private void stopTimer() {
		nsTotal += System.nanoTime() - nsStart + 1;
		msTotal += System.currentTimeMillis() - msStart;
		assert timerRunning;
		timerRunning = false;
	}

	/**
	 * Send a message.
	 * @param receiver the receiver of the message
	 * @param payload the message
	 */
	private void sendMessage(DProcess<M> receiver, M payload) {
		if ( state != State.RECEIVING ) throw new IllegalStateException("this method can only be run while receiving a message");
		state = State.SENDING;
		stopTimer();

		masterProc.enqueueMessage(this, receiver, payload);
		
		startTimer();
		state = State.RECEIVING;
	}
	
	/**
	 * Send a message.  Must be invoked by this object itself.
	 * @param receiverID the ID of the receiver of the message
	 * @param payload the message
	 */
	protected void sendMessage(int receiverID, M payload) {
		this.sendMessage(masterProc.getProcess(receiverID), payload);
	}
	
	/**
	 * Log something, including a timestamp.  Assumed to be invoked by this object itself.
	 * @param str the string to log.
	 */
	protected void log(String str) {
		if ( MasterProcess.DEBUG ) {
			long nanos = getClock();
			long ms = nanos / 1000000L;
			nanos %= 1000000L;
			System.out.println(String.format("p%03d %6d.%06d:\t%s",this.id,ms,nanos,str));
		}
	}
	
	/**
	 * Subclasses need to implement this.
	 * 
	 * @param senderID the sender of the message
	 * @param message the message received.
	 */
	protected abstract void receiveMessage(int senderID, M message);
	/**
	 * Whether or not this process is done.
	 * @return false if there is work to do yet even if no message is available; if so, receiveMessage(null,null) is called. 
	 */
	public abstract boolean isDone();
	
	@Override
	public int hashCode() {
		return id;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		DProcess<?> other = (DProcess<?>) obj;
		if (id != other.id)
			return false;
		return true;
	}

	public MasterProcess<?> getMasterProcess() {
		return this.masterProc;
	}
}
