package nl.tudelft.ewi.alg.distsim;

//  Copyright 2013 Léon Planken and James Boerkoel

/*
    This file is part of the DistSim package.

    DistSim is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DistSim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DistSim.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.PriorityQueue;

import nl.tudelft.ewi.alg.distsim.net.Immediate;
import nl.tudelft.ewi.alg.distsim.net.NetSimulator;

/**
 * The simulator of the distributed system.
 * {@link DProcess}es require a master process to register with. 
 * 
 * @param <M> The message payload to be sent around.  Can be any type.
 * 
 * @author L&eacute;on Planken
 * @author James Boerkoel Jr.
 */

public class MasterProcess<M> {
	public final static boolean DEBUG = true;
	public static final int DUMMY_PROC_ID = -1;

	static enum State {
		CREATED, RUNNING, DONE;
	}

	private final PriorityQueue<Message<M>> messageQueue;
	private final NetSimulator netsim;
	private final ArrayList<DProcess<M>> processes;

	private State state = State.CREATED;
	int[][]	sendCounters = null, receiveCounters = null;
	int[]	totalReceived = null;

	/**
	 * Construct a new MasterProcess where messages are delivered without delay and are never lost.
	 */
	public MasterProcess() {
		this(new Immediate());
	}

	/**
	 * Construct a new MasterProcess with the specified network simulator.
	 * @param netsim The network simulator.
	 */
	public MasterProcess(NetSimulator netsim) {
		this.messageQueue = new PriorityQueue<Message<M>>();
		this.netsim = netsim;
		this.processes = new ArrayList<DProcess<M>>();
	}

	/**
	 * Run the distributed algorithm.
	 */
	public void run() {
		if ( state != State.CREATED ) throw new IllegalStateException("run() may be called only once");
		
		state = State.RUNNING;
		
		assert this.sendCounters	== null;
		assert this.receiveCounters	== null;
		assert this.totalReceived	== null;
		
		final int n = this.processes.size();
		this.sendCounters		= new int[n][n];
		this.receiveCounters	= new int[n][n];
		this.totalReceived		= new int[n];

		// First, see if there are processes that want to perform some action spontaneously.
		// (This is the only way actual messages can start circulating.) 
		for ( DProcess<M> proc : processes ) {
			if ( !proc.isDone() ) {
				nullMessage(proc);
			}
		}

		// While there are messages to deliver, do so (ordered by their delivery timestamps).
		while ( !messageQueue.isEmpty() ) {
			final Message<M> m = messageQueue.poll();
			final DProcess<M> proc = m.receiver;

			if ( m.sender == null ) {
				assert m.payload == null;

				// m is a "null message".  This should only be delivered if no real messages
				// have been delivered in the meantime.
				
				// When the null message is delivered, this should be interpreted by the process as a
				// "non-blocking receive" yielding no incoming messages.
				
				if ( totalReceived[proc.id] > m.counter ) {
					// A (real) message has been delivered in the mean time; we drop this null message and carry on.
					continue;
				}
				assert proc.getClock() == m.deliveryTime : "no time elapsed after sending null message";
			} else {
				// m is a real message.
				
				++totalReceived[proc.id];
				++this.receiveCounters[m.receiver.id][m.sender.id];
			}

			proc.deliverMessage(m);
			
			if ( !proc.isDone() ) {
				// The process indicates that it has work left to do even if there are no (real) incoming messages.
				// However, if there are incoming messages, deliver these instead.
				// In other words, proc performs "non-blocking receive" which we implement by means of null messages. 
				nullMessage(proc);
			}
		}

		state = State.DONE;
	}

	public boolean isDone() {
		return state == State.DONE;
	}

	public long getParallelElapsedTime() {
		long result = Long.MIN_VALUE;
		for ( DProcess<M> proc : this.processes ){
			result = Math.max(result, proc.getClock());
		}
		return result;
	}

	public long getTotalProcessingTime() {
		return getTotalElapsedTime() - getTotalIdleTime();
	}
	
	public long getTotalElapsedTime() {
		long result = 0;
		for ( DProcess<M> proc : this.processes ) {
			result += proc.getClock();
		}
		return result;
	}

	public long getTotalIdleTime() {
		long result = 0;
		for ( DProcess<M> proc : this.processes ) {
			result += proc.getIdle();
		}
		return result;
	}

	public int[] getSendCounters(int id) {
		final int[] result = this.sendCounters[id];
		return Arrays.copyOf(result, result.length);
	}

	public int[] getReceiveCounters(int id) {
		final int[] result = this.receiveCounters[id];
		return Arrays.copyOf(result, result.length);
	}

	/**
	 * Register a process and return its id.  This method is called from DProcess's constructor.
	 * Note that MasterProcess will reserve space for max(id) processes.
	 * @param p the process to register
	 * @param id the requested ID
	 * @return id for the newly created process (equals the id parameter)
	 */
	int register(final DProcess<M> p, final int id) {
		if ( state != State.CREATED ) throw new IllegalStateException("can't register a process when run() has already been called");
		if ( id < 0 ) throw new IllegalArgumentException("negative ID");

		while ( processes.size() <= id ) {
			processes.add(null);
		}
		if ( this.processes.get(id) != null ) {
			throw new IllegalArgumentException(String.format("duplicate ID %d", id));
		}
		
		processes.set(id, p);
		return id;
	}

	void enqueueMessage(DProcess<M> sender, DProcess<M> receiver, M payload) {
		if ( state != State.RUNNING ) {
			throw new IllegalStateException("enqueueMessage() may only be called while running");
		}
		
		++this.sendCounters[sender.id][receiver.id];

		final Message<M> msg = new Message<M>(sender, receiver, payload, netsim);
		if ( msg.deliveryTime != NetSimulator.LOST ) {
			messageQueue.add(msg);
		}
	}

	private void nullMessage(DProcess<M> proc) {
		assert state == State.RUNNING;
		final Message<M> msg = new Message<M>(proc,proc.getClock(), totalReceived[proc.id]); 
		messageQueue.add(msg);
	}

	DProcess<M> getProcess(int procID) {
		try {
			DProcess<M> p = processes.get(procID);
			if ( p == null ) throw new NullPointerException();
			return p;
		} catch ( Exception e ) {
			throw new IllegalArgumentException("illegal process id");
		}
	}
	
	public List<DProcess<M>> processes() {
		return Collections.unmodifiableList(this.processes);
	}
}
