package nl.tudelft.ewi.alg.distsim.example;

//  Copyright 2013 Léon Planken and James Boerkoel

/*
    This file is part of the DistSim package.

    DistSim is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DistSim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DistSim.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.Arrays;
import java.util.Random;

import nl.tudelft.ewi.alg.distsim.DProcess;
import nl.tudelft.ewi.alg.distsim.MasterProcess;

/**
 * These distributed processes each start off with a message of their own.
 * This message is a "chain letter" of sorts and must be passed around randomly
 * a couple of times, until it eventually winds up again with the originator.
 * 
 * @author L@eacute;on Planken
 */
public class MyProc extends DProcess<ChainLetter> {
	public static final String[] MESSAGES = new String[] { "ALPHA", "BRAVO", "CHARLIE", "DELTA", "ECHO", "FOXTROT", "GOLF", "HOTEL" };
	public static final int COMPUTATION_TIME = 1000;
	
	private final Random rnd;
	private final int[] procs;
	
	/**
	 * My chain letter (if I didn't send it off)
	 */
	private ChainLetter mymsg;
	
	/**
	 * Whether I have some action I'd like to perform spontaneously.
	 */
	boolean done = false;
	
	public MyProc(MasterProcess<ChainLetter> master, int id, int processcount) {
		super(master, id);
		
		this.rnd = new Random();
		
		this.procs = new int[processcount];
		Arrays.fill(this.procs, MasterProcess.DUMMY_PROC_ID);
		
		this.mymsg = new ChainLetter(MESSAGES[this.id], this.id);
		// add some random recipients for the chain letter
		int last = this.id;
		for ( int i = 0; i < rnd.nextInt(5) + 5; ++i ) {
			int next;
			do {
				next = rnd.nextInt(Example.N);
			} while ( next == last );
			this.mymsg.enqueueReceiverID(next);
			last = next;
		}

		// make sure the message gets back to this process in the end
		if ( last != this.id ) {
			this.mymsg.enqueueReceiverID(this.id);
		}
	}

	@Override
	public boolean isDone() { return done; }

	@Override
	protected void receiveMessage(int senderID, ChainLetter msg) {
		if ( msg == null ) {
			assert senderID == MasterProcess.DUMMY_PROC_ID : "only Master sends null messages";
			assert !done : "if I advertised myself as 'done', I wouldn't get this opportunity";

			// No message is incoming, but I get to perform some action spontaneously.
			spontaneousAction();
		} else {
			// An actual message is incoming.
			handleMsg(senderID, msg);
		}
	}

	private void handleMsg(int senderID, ChainLetter msg) {
		if ( !msg.queueEmpty() ) {
			log(String.format("Got chain letter %s from p%d; queue: %s", msg.text, senderID, msg.queueContents()));
			
			// This chain letter still needs to be forwarded.
			int id = msg.pollReceiverID();
			sendMessage(this.procs[id], msg);
		} else {
			assert msg.originatorID == this.id : "the chain letter is eventually returned to me";
			this.done = true;
			log(String.format("I got my chain letter %s back (from p%d)!", msg.text, senderID));
		}
	}

	private void spontaneousAction() {
		if ( this.mymsg != null ) {
			// I still have my chain letter!  Let's send it off.
			log(String.format("Sending off chain letter %s; queue: %s", this.mymsg.text, this.mymsg.queueContents()));
			sendMessage(this.id, this.mymsg);
			this.mymsg = null;
		} else {
			// I can perform some spontaneous action here. Simulating computation of my own by sleeping.
			final int time = this.rnd.nextInt(COMPUTATION_TIME);
			log(String.format(String.format("Doing some computation for %dms", time)));
			try {
				Thread.sleep(time);
			} catch ( InterruptedException e ) {}
		}
	}

	public void setProcs(int[] processIDs) {
		System.arraycopy(processIDs, 0, this.procs, 0, processIDs.length);
	}
}
