package nl.tudelft.ewi.alg.distsim.net;

//  Copyright 2013 Léon Planken and James Boerkoel

/*
    This file is part of the DistSim package.

    DistSim is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DistSim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DistSim.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.ArrayList;
import java.util.Random;

/**
 * Simulator for a non-lossy network with uniform random latency between two bounds.
 * @author L&eacute;on Planken
 */

public class RandomDelay implements NetSimulator {
	private ArrayList<ArrayList<Long>> lastSent;
	private final Random rnd;
	private final int maxMS;
	private final int minMS;
	
	/**
	 * Construct a new instance with the specified bounds. 
	 * @param rnd Random number generator
	 * @param minMS Minimum delay in milliseconds
	 * @param maxMS Maximum delay in milliseconds
	 */
	public RandomDelay(Random rnd, int minMS, int maxMS) {
		if ( maxMS < minMS ) throw new IllegalArgumentException("max < min");
		this.lastSent = new ArrayList<ArrayList<Long>>();
		this.rnd = rnd;
		this.maxMS = maxMS;
		this.minMS = minMS;
	}

	/** {@inheritDoc} */
	public long send(int senderID, int receiverID, long sendTime) {
		check(senderID); check(receiverID);
		long last = lastSent.get(senderID).get(receiverID);
		long delay = rnd.nextInt(maxMS - minMS) + minMS;
		delay *= 1000;
		delay *= 1000;
		long retval = sendTime + delay;
		if ( retval < last ) return last;
		lastSent.get(senderID).set(receiverID, retval);
		return retval;
	}

	private void check(int idx) {
		if ( lastSent.size() <= idx ) {
			for ( ArrayList<Long> l : lastSent ) {
				l.ensureCapacity(idx+1);
				grow(l,idx);
			}
			lastSent.ensureCapacity(idx + 1);
			while ( lastSent.size() <= idx ) {
				ArrayList<Long> foo = new ArrayList<Long>(idx+1);
				grow(foo,idx);
				lastSent.add(foo);
			}
		}
	}

	private void grow(ArrayList<Long> l, int idx) {
		while ( l.size() <= idx ) {
			l.add(Long.MIN_VALUE);
		}
	}
}
